# bap
**B**ead-based sc**A**TAC-seq data **P**rocessing

## About
This is an integrated `python` (>= 3.6) and `R` package that
performs analysis and identification of barcode multiplets and processes
data files for downstream analyses for droplet-based single-cell ATAC-seq data. 

## Documentation

[See the bap wiki page](https://github.com/caleblareau/bap/wiki) for details and FAQs about bap.



